@@ -1,10 +1,13 @@
+from collections import OrderedDict as odict
+
 import numpy as np
-from astropy.utils import OrderedDict as odict
 from astropy.io import fits
 from astropy.table import Table, vstack
+import yaml
 
 __all__ = ['read_snana_ascii', 'read_snana_fits', 'read_snana_simlib']
 
+
 def read_snana_fits(head_file, phot_file, snids=None, n=None):
     """Read the SNANA FITS format: two FITS files jointly representing
     metadata and photometry for a set of SNe.
@@ -35,10 +38,10 @@ def read_snana_fits(head_file, phot_file, snids=None, n=None):
 
     Examples
     --------
-    >>> sne = read_snana_fits('HEAD.fits', 'PHOT.fits')  # doctest: +SKIP
-    >>> for sn in sne:                                   # doctest: +SKIP
-    ...     sn.meta  # Metadata in an OrderedDict.       # doctest: +SKIP
-    ...     sn['MJD']  # MJD column                      # doctest: +SKIP
+    >>> sne = read_snana_fits('HEAD.fits', 'PHOT.fits')
+    >>> for sn in sne:
+    ...     sn.meta  # Metadata in an OrderedDict.
+    ...     sn['MJD']  # MJD column
 
     """
 
@@ -64,11 +67,11 @@ def read_snana_fits(head_file, phot_file, snids=None, n=None):
             raise RuntimeError('Specific snids requested, but head file does'
                                ' not contain SNID column')
         idx = []
-        for snid in snids: 
+        for snid in snids:
             i = np.flatnonzero(head_data['SNID'] == snid)
             if len(i) != 1:
-                raise RuntimeError('Unique snid requested, but there are {0:d} '
-                                   'matching entries'.format(len(i)))
+                raise RuntimeError('Unique snid requested, but there are '
+                                   '{0:d} matching entries'.format(len(i)))
             idx.append(i[0])
     elif snids is None:
         idx = range(n)
@@ -80,12 +83,12 @@ def read_snana_fits(head_file, phot_file, snids=None, n=None):
     for i in idx:
         meta = odict(zip(head_data.dtype.names, head_data[i]))
 
-        j0 = head_data['PTROBS_MIN'][i] - 1 
+        j0 = head_data['PTROBS_MIN'][i] - 1
         j1 = head_data['PTROBS_MAX'][i]
-	data = phot_data[j0:j1]
+        data = phot_data[j0:j1]
         if 'FLT' in data.dtype.names:
             data['FLT'][:] = np.char.strip(data['FLT'])
-	sne.append(Table(data, meta=meta, copy=False))
+        sne.append(Table(data, meta=meta, copy=False))
 
     return sne
 
@@ -141,12 +144,14 @@ def read_snana_ascii(fname, default_tablename=None):
     * Any other keywords anywhere in the file are treated as metadata. The
       first string after the keyword is treated as the value for that keyword.
     * **Note:** Newlines are treated as equivalent to spaces; they do not
-      indicate a new row.
+      indicate a new row. This is necessary because some SNANA-format files
+      have multiple metadata on a single row or single table rows split over
+      multiple lines, making newline characters meaningless.
 
     Examples
     --------
 
-    >>> from StringIO import StringIO  # StringIO behaves like a file
+    >>> from io import StringIO  # StringIO behaves like a file
     >>> f = StringIO('META1: a\\n'
     ...              'META2: 6\\n'
     ...              'NVAR_SN: 3\\n'
@@ -165,7 +170,7 @@ def read_snana_ascii(fname, default_tablename=None):
 
     >>> tables['SN']
     <Table rows=2 names=('A','B','C')>
-    array([(1, 2.0, 'x'), (4, 5.0, 'y')], 
+    array([(1, 2.0, 'x'), (4, 5.0, 'y')],
           dtype=[('A', '<i8'), ('B', '<f8'), ('C', 'S1')])
 
 
@@ -179,15 +184,15 @@ def read_snana_ascii(fname, default_tablename=None):
 
     it can be read by supplying a default table name:
 
-     >>> meta, tables = read_snana_ascii(f, default_tablename='SN')  # doctest: +SKIP
+    >>> meta, tables = read_snana_ascii(f, default_tablename='SN')
 
     """
 
-    meta = odict() # initialize structure to hold metadata.
-    tables = {} # initialize structure to hold data.
+    meta = odict()  # initialize structure to hold metadata.
+    tables = {}  # initialize structure to hold data.
 
-    if isinstance(fname, basestring):
-        fh = open(fname, 'U')
+    if isinstance(fname, str):
+        fh = open(fname, 'r')
     else:
         fh = fname
     words = fh.read().split()
@@ -203,7 +208,7 @@ def read_snana_ascii(fname, default_tablename=None):
         if word.startswith('NVAR'):
             nvar = int(words[i + 1])
 
-            #Infer table name. The name will be used to designate a data row.
+            # Infer table name. The name will be used to designate a data row.
             if '_' in word:
                 pos = word.find('_') + 1
                 tablename = word[pos:].rstrip(':')
@@ -211,7 +216,7 @@ def read_snana_ascii(fname, default_tablename=None):
                 tablename = default_tablename
             else:
                 raise ValueError(
-                    'table name must be given as part of NVAR keyword so '
+                    'Table name must be given as part of NVAR keyword so '
                     'that rows belonging to this table can be identified. '
                     'Alternatively, supply the default_tablename keyword.')
             table = odict()
@@ -221,7 +226,7 @@ def read_snana_ascii(fname, default_tablename=None):
 
         # If the word starts with 'VARNAMES', the following `nvar` words
         # define the column names of the table.
-        elif word.startswith('VARNAMES'):
+        elif word.startswith('VARNAMES') or word.startswith('VARLIST'):
 
             # Check that nvar is defined and that no column names are defined
             # for the current table.
@@ -238,7 +243,7 @@ def read_snana_ascii(fname, default_tablename=None):
             for j, colname in enumerate(table.keys()):
                 table[colname].append(words[i + 1 + j])
             i += nvar + 1
-        
+
         # Otherwise, we are reading metadata or some comment
         # If the word ends with ":", it is metadata.
         elif word[-1] == ':':
@@ -252,7 +257,7 @@ def read_snana_ascii(fname, default_tablename=None):
                     except ValueError:
                         val = words[i + 1]
                 meta[name] = val
-            else: 
+            else:
                 meta[name] = None
             i += 2
         else:
@@ -262,7 +267,7 @@ def read_snana_ascii(fname, default_tablename=None):
     # All values in each column are currently strings. Convert to int or
     # float if possible.
     for table in tables.values():
-        for colname, values in table.iteritems():
+        for colname, values in table.items():
             try:
                 table[colname] = [int(val) for val in values]
             except ValueError:
@@ -295,8 +300,7 @@ def read_snana_ascii_multi(fnames, default_tablename=None):
     Examples
     --------
     >>> tables = read_snana_ascii_multi(['data1.txt', 'data1.txt'])
-    ... # doctest: +SKIP
-   
+
     """
 
     alltables = {}
@@ -304,7 +308,7 @@ def read_snana_ascii_multi(fnames, default_tablename=None):
         meta, tables = read_snana_ascii(fname,
                                         default_tablename=default_tablename)
 
-        for key, table in tables.iteritems():
+        for key, table in tables.items():
             if key in alltables:
                 alltables[key].append(table)
             else:
@@ -315,6 +319,7 @@ def read_snana_ascii_multi(fnames, default_tablename=None):
 
     return alltables
 
+
 def _parse_meta_from_line(line):
     """Return dictionary from key, value pairs on a line. Helper function for
     snana_read_simlib."""
@@ -363,7 +368,8 @@ def read_snana_simlib(fname):
     Returns
     -------
     meta : `OrderedDict`
-        Global meta data, not associated with any one LIBID.
+        Global meta data, not associated with any one LIBID. If
+        DOCANA is present, it is stored in `meta['DOCUMENTATION']`.
     observation_sets : `OrderedDict` of `astropy.table.Table`
         keys are LIBIDs, values are observation sets.
 
@@ -378,29 +384,32 @@ def read_snana_simlib(fname):
       the keyword. These are (1) MJD, (2) IDEXPT, (3) FLT, (4) CCD GAIN,
       (5) CCD NOISE, (6) SKYSIG, (7) PSF1, (8) PSF2, (9) PSF 2/1 RATIO,
       (10) ZPTAVG, (11) ZPTSIG, (12) MAG.
+    * Column (2) may represent co-added observations in a '111*1' format. In
+      this case, the 'IDEXPT' column is split at the '*' into 'IDEXPT' and
+      'NEXPOSE'
     * Other lines inside a 'LIBID:'/'END_LIBID:' pair are treated as metadata
       for that LIBID.
     * Any other keywords outside a 'LIBID:'/'END_LIBID:' pair are treated
       as global header keywords and are returned in the `meta` dictionary.
-    
+
     Examples
     --------
-    >>> meta, obs_sets = read_snana_simlib('filename') # doctest: +SKIP
+    >>> meta, obs_sets = read_snana_simlib('filename')
 
     The second object is a dictionary of astropy Tables indexed by LIBID:
 
-    >>> obs_sets.keys()  # doctest: +SKIP
-    [0, 1, 2, 3, 4] 
+    >>> obs_sets.keys()
+    [0, 1, 2, 3, 4]
 
     Each table (libid) has metadata:
 
-    >>> obs_sets[0].meta  # doctest: +SKIP
+    >>> obs_sets[0].meta
     OrderedDict([('LIBID', 0), ('RA', 52.5), ('DECL', -27.5), ('NOBS', 161),
                  ('MWEBV', 0.0), ('PIXSIZE', 0.27)])
 
     Each table has the following columns:
 
-    >>> obs_sets[0].colnames  # doctest: +SKIP
+    >>> obs_sets[0].colnames
     ['SEARCH', 'MJD', 'IDEXPT', 'FLT', 'CCD_GAIN', 'CCD_NOISE', 'SKYSIG',
      'PSF1', 'PSF2', 'PSFRATIO', 'ZPTAVG', 'ZPTSIG', 'MAG']
 
@@ -409,14 +418,18 @@ def read_snana_simlib(fname):
     from astropy.table import Table
 
     COLNAMES = ['SEARCH', 'MJD', 'IDEXPT', 'FLT', 'CCD_GAIN', 'CCD_NOISE',
-                'SKYSIG', 'PSF1', 'PSF2', 'PSFRATIO', 'ZPTAVG', 'ZPTSIG', 'MAG']
-    SPECIAL = ['FIELD', 'TELESCOPE', 'PIXSIZE'] # not used yet... if present in 
-                                                # header, add to table.
+                'SKYSIG', 'PSF1', 'PSF2', 'PSFRATIO', 'ZPTAVG', 'ZPTSIG',
+                'MAG']
+
+    # Not used yet... if present in header, add to table.
+    SPECIAL = ['FIELD', 'TELESCOPE', 'PIXSIZE']
 
     meta = odict()  # global metadata
-    observation_sets = odict() # dictionary of tables indexed by LIBID
+    observation_sets = odict()  # dictionary of tables indexed by LIBID
 
     reading_obsset = False
+    reading_docana = False
+    docana = ''
     with open(fname, 'r') as infile:
         for line in infile.readlines():
 
@@ -427,40 +440,86 @@ def read_snana_simlib(fname):
 
             # split on spaces.
             words = line.split()
-            if len(words) == 0: continue
+            if len(words) == 0:
+                continue
 
-            # are we currently reading an obsset?
+            # If we're not currently reading an obs set, check if this line
+            # is the start of one. If it isn't, update the global metadata.
             if not reading_obsset:
                 if line[0:6] == 'LIBID:':
                     reading_obsset = True
                     current_meta = _parse_meta_from_line(line)
                     current_data = odict([(key, []) for key in COLNAMES])
                 else:
-                    meta.update(_parse_meta_from_line(line))
+                    # If we're currently reading an documentation block, add
+                    # it to the DOCANA string.
+                    # If we're not, check if this line is the start of one,
+                    # then update the global metadata.
+                    if reading_docana:
+                        if line[0:18] == 'DOCUMENTATION_END:':
+                            reading_docana = False   # And then skip line
+                            meta.update(
+                                odict({'DOCUMENTATION':
+                                      yaml.safe_load(docana)})
+                                )
+                            docana = ''
+                        else:
+                            docana += line + '\n'
+                    else:
+                        if line[0:14] == 'DOCUMENTATION:':
+                            reading_docana = True    # And then skip line
+                        else:
+                            meta.update(_parse_meta_from_line(line))
 
+            # If we are currently reading an obsset...
             else:
+                # Check for the explicit end of the obs set.
                 if line[0:10] == 'END_LIBID:':
                     reading_obsset = False
                     observation_sets[current_meta['LIBID']] = \
                         Table(current_data, meta=current_meta)
+
+                # Sometimes there's not an explicit end, but the next one
+                # starts anyway.
+                elif line[0:6] == 'LIBID:':
+                    observation_sets[current_meta['LIBID']] = \
+                        Table(current_data, meta=current_meta)
+                    current_meta = _parse_meta_from_line(line)
+                    current_data = odict([(key, []) for key in COLNAMES])
+
+                # Otherwise, read the line into the current obs set.
                 elif line[0:2] in ['S:', 'T:']:
                     words = line.split()
-                    for colname, val in [
-                        ('SEARCH', words[0] == 'S:'),
-                        ('MJD', float(words[1])),
-                        ('IDEXPT', int(words[2])),
-                        ('FLT', words[3]),
-                        ('CCD_GAIN', float(words[4])),
-                        ('CCD_NOISE', float(words[5])),
-                        ('SKYSIG', float(words[6])),
-                        ('PSF1', float(words[7])),
-                        ('PSF2', float(words[8])),
-                        ('PSFRATIO', float(words[9])),
-                        ('ZPTAVG', float(words[10])),
-                        ('ZPTSIG', float(words[11])),
-                        ('MAG', float(words[12]))]:
+                    colnames = ['SEARCH', 'MJD', 'IDEXPT', 'FLT', 'CCD_GAIN',
+                                'CCD_NOISE', 'SKYSIG', 'PSF1', 'PSF2',
+                                'PSFRATIO', 'ZPTAVG', 'ZPTSIG', 'MAG']
+                    values = [words[0] == 'S:', float(words[1]), None,
+                              words[3], float(words[4]), float(words[5]),
+                              float(words[6]), float(words[7]),
+                              float(words[8]), float(words[9]),
+                              float(words[10]), float(words[11]),
+                              float(words[12])]
+                    if "*" in words[2]:
+                        # 'IDEXPT' can be a co-add of the form: '2063*2'
+                        # re-process assuming co-added expsoures
+                        # ('IDEXPT' -> 'IDEXPT', 'NEXPOSE' )
+                        colnames.insert(3, "NEXPOSE")
+                        if 'NEXPOSE' not in current_data:
+                            # add an empty list only on the first data line
+                            current_data['NEXPOSE'] = []
+                        values[2] = int(words[2].split('*')[0])
+                        values.insert(3, int(words[2].split('*')[1]))
+                    else:
+                        values[2] = int(words[2])
+                    for colname, val in zip(colnames, values):
                         current_data[colname].append(val)
                 else:
                     current_meta.update(_parse_meta_from_line(line))
 
+    # At the end, check for the case where there's not an explicit end
+    # to the last obs set:
+    if reading_obsset:
+        observation_sets[current_meta['LIBID']] = \
+            Table(current_data, meta=current_meta)
+
     return meta, observation_sets
